@@ -8,32 +8,45 @@ check_has_converged <- function(old_value, new_value, tolerance) {
   }
 }
 
+
+# Make matrix symmetric to check for too sparse clusters
+make_symmetric <- function(g){
+  g_symmetric <- as(g, "TsparseMatrix")
+  g_symmetric <- sparseMatrix(i=c(g_symmetric@i+1, g_symmetric@j+1),
+                              j=c(g_symmetric@j+1, g_symmetric@i+1),
+                              dims=g_symmetric@Dim)
+  g_symmetric <- as(g_symmetric, "TsparseMatrix")
+  g_symmetric <- as(g_symmetric, "dMatrix")
+  return(g_symmetric)
+}
+
 # Wrapper function that conducts clustering.
 MM_wrapper <-
   function(network,
            formula,
-           n_clusters,
+           n_blocks,
            n_MM_step_max,
            tol_MM_step,
            min_size = 2,
-           initialization_method = 1,
+           initialization = 1,
            use_infomap_python = FALSE,
            virtualenv_python = "r-bigergm",
            seed_infomap = NULL,
-           initialized_cluster_data = NULL,
            clustering_with_features = FALSE,
            list_multiplied_feature_matrices = NULL,
            verbose = 0,
            weight_for_initialization = 1000,
            compute_pi = FALSE,
            check_alpha_update = FALSE,
-           check_block_membership = FALSE,
+           check_blocks = FALSE,
            MM_restart_object = NULL,
            minAlpha = 1e-6,
            cache,
            ...) {
     n_nodes <- as.integer(network$gal$n)
-    is_undirected <- !network$gal$directed
+    
+    is_directed <- network$gal$directed
+    
     
     # Cache eigenvectors computation in disk to free memory for other computations
     if (is.null(cache)) {
@@ -46,37 +59,34 @@ MM_wrapper <-
     if (is.null(MM_restart_object)) {
       # Step 0: Initialization
       n_nodes <- as.integer(network$gal$n)
-      is_undirected <- !network$gal$directed
-      
       if (verbose > 0) message("Converting network to edgelist...")
       edgelist <- network::as.edgelist(network)
       
       if (verbose > 0) message("Converting edgelist to sparse matrix...")
-      g <- Matrix::sparseMatrix(i = edgelist[, 1], j = edgelist[, 2], x = 1, dims = c(n_nodes, n_nodes), symmetric = is_undirected)
       
+      g <- Matrix::sparseMatrix(i = edgelist[, 1], j = edgelist[, 2], x = 1, dims = c(n_nodes, n_nodes), symmetric = !is_directed)
       ## Calculate network statistics needed for variational approximation of p(Z|X)
       if (verbose > 0) message("\nStep 1: Initialize z")
       
-      # Step 1: Get initial estimate of Z memberships
-      # If you already have an initialized clustering result, start the MM iterations from the given clustering result.
-      if (!is.null(initialized_cluster_data)) {
-        z_memb_init <- factor(initialized_cluster_data, levels = 1:n_clusters)
-        z_memb <- check_clusters_sparse(z_memb_init, g, n_clusters, eigenvectors_sparse_fn, min_size, verbose = verbose)
-      }
-      
       # Start cluster initialization
-      else if (initialization_method == 1) # Infomap
+      if(length(initialization) >1){
+        if(length(initialization) == network$gal$n){
+          z_memb <- z_memb_init <- factor(match(initialization, unique(initialization)), levels = 1:n_blocks)
+        } else {
+          stop("Initialization not recognized.")
+        }
+      } else if (initialization == "infomap") # Infomap
       {
         if (verbose > 0) message("Using Infomap to initialize clustering step...")
         
         if(use_infomap_python){
           
           # first check if an env is available
-          if(!reticulate::py_available())
+          if(!reticulate::virtualenv_exists(virtualenv_python))
           {
-            message("No Python Environemt available. Use check_and_install() ",
-                    "to install recommended environment.")
-            invisible(return(NULL))
+            stop("No Python Environemt available. Use py_dep() ",
+                 "to install recommended environment.")
+            
           }
           reticulate::use_virtualenv(virtualenv_python, required = TRUE)
           have_infomap <- reticulate::py_module_available("infomap")
@@ -84,30 +94,71 @@ MM_wrapper <-
           
           if (have_numpy + have_infomap != 2) {
             use_infomap_python <- F
-            message("Since either Numpy or Infomap is not available in Pythin the igraph implementation is used.")
-          }
+            
+            message("Since either Numpy or Infomap is not available in Pythin the igraph implementation is used (see py_dep).")
+          }  
         }
         
-        if (use_infomap_python) {
+        if(use_infomap_python){
           python_path <- system.file("python", package = "bigergm")
           opt <- reticulate::import_from_path("infomap_R", path = python_path)
           
-          z_memb_init <- factor(opt$infomap_python(matrix = edgelist, n_clusters = n_clusters), levels = 1:n_clusters)
-          z_memb <- check_clusters_sparse(z_memb_init, g, n_clusters, eigenvectors_sparse_fn, min_size, verbose = verbose)
+          if(is_directed){
+            z_memb_init <- factor(opt$infomap_python_directed(matrix = edgelist,n_clusters = n_blocks), levels = 1:n_blocks)
+            
+            # Make matrix symmetric to check for too sparse clusters
+            g_symmetric <- make_symmetric(g)
+            z_memb <- check_clusters_sparse(z_memb_init, g_symmetric, n_blocks, eigenvectors_sparse_fn, min_size, verbose = verbose)  
+          } else {
+            z_memb_init <- factor(opt$infomap_py$infomap_python(matrix = edgelist,n_clusters = n_blocks), levels = 1:n_blocks)
+            z_memb <- check_clusters_sparse(z_memb_init, g, n_blocks, eigenvectors_sparse_fn, min_size, verbose = verbose)   
+          }
         } else {
           # Use igraph's infomap
           set.seed(seed_infomap)
           b <- igraph::cluster_infomap(intergraph::asIgraph(network))
-          z_memb_init <- factor(b$membership, levels = 1:n_clusters)
-          z_memb <- check_clusters_sparse(z_memb_init, g, n_clusters, eigenvectors_sparse_fn, min_size, verbose = verbose)
+          z_memb_init <- factor(b$membership, levels = 1:n_blocks)
+          
+          if(is_directed){
+            
+            g_symmetric <- make_symmetric(g)
+            z_memb <- check_clusters_sparse(z_memb = z_memb_init,network =  g_symmetric, 
+                                            n_blocks = n_blocks, eigenvectors_fn = eigenvectors_sparse_fn, 
+                                            min_size = min_size, verbose = verbose)  
+          } else {
+            z_memb <- check_clusters_sparse(z_memb_init, g, n_blocks, 
+                                            eigenvectors_sparse_fn,
+                                            min_size, verbose = verbose)   
+          }
         }
-      } else if (initialization_method == 2) {
+        
+      }
+      else if (initialization == "random") {
         if (verbose > 0) message("Initializing with uniformly distributed clusters...")
-        z_memb <- z_memb_init <- factor(sample.int(n_clusters, size = n_nodes, replace = TRUE))
-      } else # Spectral clustering
+        z_memb <- z_memb_init <- factor(sample.int(n_blocks, size = n_nodes, replace = TRUE))
+      } else if (initialization == "spectral") # Spectral clustering
       {
-        z_memb <- spec_clust_sparse(g, n_clusters, eigenvectors_sparse_fn)
+        z_memb <- spec_clust_sparse(g, n_blocks, eigenvectors_sparse_fn)
         z_memb_init <- z_memb
+      } else if (initialization == "walktrap") 
+      {
+        # Use igraph's infomap
+        set.seed(seed_infomap)
+        b <- igraph::cluster_walktrap(intergraph::asIgraph(network))
+        z_memb_init <- factor(b$membership, levels = 1:n_blocks)
+  
+        if(is_directed){
+          g_symmetric <- make_symmetric(g)
+          z_memb <- check_clusters_sparse(z_memb = z_memb_init,network =  g_symmetric, 
+                                          n_blocks = n_blocks, eigenvectors_fn = eigenvectors_sparse_fn, 
+                                          min_size = min_size, verbose = verbose)  
+        } else {
+          z_memb <- check_clusters_sparse(z_memb_init, g, n_blocks, 
+                                          eigenvectors_sparse_fn,
+                                          min_size, verbose = verbose)   
+        }
+      } else {
+        stop("Initialization not recognized.")
       }
       
       # Step 2: Find A(Z=z) ~ P(Z=z|X=x)
@@ -119,13 +170,11 @@ MM_wrapper <-
       n_blocks <- length(block_sizes)
       
       # Initialize alpha
-      if (verbose > 0) message("Initializing alpha...")
-      alpha <- matrix(1, n_nodes, n_clusters)
-      if (verbose > 0) message("Replacing zeros in alpha...")
+      if (verbose > 0) message("Initializing posterior estimates")
+      alpha <- matrix(1, n_nodes, n_blocks)
       for (i in which(!is.na(z_memb))) {
         alpha[i, z_memb[i]] <- weight_for_initialization
       }
-      if (verbose > 0) message("Normalizing alpha...")
       alpha <- alpha / Matrix::rowSums(alpha)
       
       # Keep track of alpha, z, changes in alpha, and the lower bound
@@ -136,7 +185,7 @@ MM_wrapper <-
       if (check_alpha_update) {
         list_alpha[[1]] <- alpha
       }
-      if (check_block_membership) {
+      if (check_blocks) {
         list_z[[1]] <- z_memb
       }
       # Set the counter of MM iteration
@@ -144,17 +193,16 @@ MM_wrapper <-
     }
     # If restart_object is given, restore the values.
     else {
-      if (verbose > 0) message("Making alpha matrix dense...")
       alpha <- as.matrix(MM_restart_object$alpha)
-      alpha[alpha == 0] <- minAlpha
+      alpha[alpha < minAlpha] <- minAlpha
       list_alpha <- MM_restart_object$list_alpha
       z_memb_init <- MM_restart_object$z_memb_init
       list_z <- MM_restart_object$list_z
       change_in_alpha <- MM_restart_object$change_in_alpha
       lower_bound <- MM_restart_object$lower_bound
       counter_e_step <- MM_restart_object$counter_e_step
-      g <- MM_restart_object$adjacency_matrix
-      n_blocks <- n_clusters
+      g <- MM_restart_object$adjacency_matrix*1
+      n_blocks <- n_blocks
     }
     
     # Get a sparse feature adjacency matrix if clustering with features
@@ -162,32 +210,32 @@ MM_wrapper <-
       # If a list of feature matrices is given from outside, skip this step.
       if (!is.null(list_multiplied_feature_matrices)) {
         if (verbose > 0) {
-          message("Skipped forming features matrix")
+          message("Skipped preprocessing")
         }
       } else {
         if (verbose > 0) {
-          message("Forming features matrix")
-        }
-        list_feature_matrices <- get_list_sparse_feature_adjmat(network, formula)
-        if (verbose > 0) {
-          message("Got features matrix")
+          message("Starting preprocessing")
         }
+        list_feature_matrices <- get_features(network, formula)$list_sparse_feature_adjmat
       }
+    } else {
     }
     
     # Get multiplied feature matrices if clustering with features
     if (clustering_with_features) {
-      if (is.null(list_multiplied_feature_matrices)) {
-        if (verbose > 0) {
-          message("Getting multiplied matrices used for E step")
+      if (is.null(list_multiplied_feature_matrices) ) {
+        if(is_directed){
+          g <- as(g, "TsparseMatrix")
+          list_multiplied_feature_matrices <- get_elementwise_multiplied_matrices_R(g, list_feature_matrices)
+        } else {
+          list_multiplied_feature_matrices <- get_elementwise_multiplied_matrices_R(g, list_feature_matrices)
+          if (verbose > 0) {
+            message("Finished preprocessing")
+          }
+          rm(list_feature_matrices)
         }
-        list_multiplied_feature_matrices <- get_elementwise_multiplied_matrices(g, list_feature_matrices)
-        if (verbose > 0) {
-          message("Got multiplied matrices")
-        }
-        rm(list_feature_matrices)
       } else {
-        message("Skipped getting multiplied matrices")
+        message("Skipped preprocessing")
       }
     }
     
@@ -204,6 +252,8 @@ MM_wrapper <-
     
     # This is for internal use
     inner_counter <- 0
+    # This just changes the type of sparse matrices from "nsCMatrix" to "dsCMatrix"
+    list_multiplied_feature_matrices <- lapply(list_multiplied_feature_matrices, function(x) x*1)
     
     if (clustering_with_features) {
       # MM iterations start
@@ -224,14 +274,12 @@ MM_wrapper <-
         
         # Update alpha
         alpha_LB <-
-          run_MM_with_features(
-            as.integer(n_nodes),
-            as.integer(n_clusters),
-            as.double(gamma),
-            list_multiplied_feature_matrices,
-            alpha,
-            verbose
-          )
+          run_MM_with_features(numOfVertices = n_nodes,
+                               numOfClasses = n_blocks,
+                               alpha = gamma,
+                               tau = alpha,
+                               list_multiplied_feature_adjmat = list_multiplied_feature_matrices, 
+                               verbose = verbose,directed = is_directed)
         
         alpha <- alpha_LB[[1]]
         gamma <- colMeans(alpha)
@@ -253,15 +301,15 @@ MM_wrapper <-
         }
         
         # If you would like to keep track on block memberships over MM iterations:
-        if (check_block_membership) {
-          z <- factor(apply(alpha, 1, which.max), levels = 1:n_clusters)
+        if (check_blocks) {
+          z <- factor(apply(alpha, 1, which.max), levels = 1:n_blocks)
           list_z[[counter_e_step + 1]] <- z
         }
         
         if (verbose > 0) {
           iter_end <- Sys.time()
           message(glue::glue("MM algorithm iteration {counter_e_step} finished at {iter_end}"))
-          message(glue::glue("MM iterations {counter_e_step} took {signif(difftime(iter_end, iter_start, units = 'mins'), digits = 3)} minutes."))
+          message(glue::glue("MM iteration {counter_e_step} took {signif(difftime(iter_end, iter_start, units = 'mins'), digits = 3)} minutes."))
         }
         
         if (inner_counter %% 5) {
@@ -281,11 +329,8 @@ MM_wrapper <-
         message(glue::glue("Total MM iterations: {counter_e_step}"))
       }
       
-      # Compute pi if necessary
-      Pi <- NULL
-      if (compute_pi) {
-        Pi <- compute_pi_with_features(n_nodes, n_clusters, list_multiplied_feature_matrices, alpha)
-      }
+      # Compute pi at the end of the MM iterations
+      Pi <- compute_pi_with_features(n_nodes, n_blocks, list_multiplied_feature_matrices, alpha)
     } else { # If clustering without features:
       list_multiplied_feature_matrices <- NULL
       # MM iterations start
@@ -311,7 +356,8 @@ MM_wrapper <-
             gamma,
             alpha,
             g,
-            verbose
+            verbose,
+            is_directed
           )
         
         # \alpha and \gamma
@@ -335,8 +381,8 @@ MM_wrapper <-
         }
         
         # If you would like to keep track on block memberships over MM iterations:
-        if (check_block_membership) {
-          z <- factor(apply(alpha, 1, which.max), levels = 1:n_clusters)
+        if (check_blocks) {
+          z <- factor(apply(alpha, 1, which.max), levels = 1:n_blocks)
           list_z[[counter_e_step + 1]] <- z
         }
         
@@ -365,15 +411,12 @@ MM_wrapper <-
       }
       
       # Compute pi if necessary
-      Pi <- NULL
-      if (compute_pi) {
-        Pi <- (t(alpha) %*% g %*% alpha) / compute_sumTaus(n_nodes, n_blocks, alpha)
-      }
+      Pi <- (t(alpha) %*% g %*% alpha) / compute_sumTaus(n_nodes, n_blocks, alpha)
     }
     
     # Estimate the block membership
     z_memb <-
-      factor(apply(alpha, 1, which.max), levels = 1:n_clusters)
+      factor(apply(alpha, 1, which.max), levels = 1:n_blocks)
     # Keep the final clustering result before check_cluster_sparse()
     z_memb_final_before_kmeans <- z_memb
     
@@ -382,10 +425,22 @@ MM_wrapper <-
     alpha <- as(alpha, "dgCMatrix")
     
     # Check bad clusters
-    z_memb_final <-
-      factor(check_clusters_sparse(z_memb, g, n_clusters, eigenvectors_sparse_fn, min_size, verbose = verbose),
-             levels = 1:n_clusters
-      )
+    
+    if(is_directed){
+      if(!exists("g_symmetric")) {
+        g_symmetric <- make_symmetric(g)
+      }
+      z_memb_final <-
+        factor(check_clusters_sparse(z_memb, g_symmetric, n_blocks, eigenvectors_sparse_fn, min_size, verbose = verbose),
+               levels = 1:n_blocks
+        )    
+    } else {
+      z_memb_final <-
+        factor(check_clusters_sparse(z_memb, g*1, n_blocks, eigenvectors_sparse_fn, min_size, verbose = verbose),
+               levels = 1:n_blocks
+        )
+    }
+    
     
     # Return the output
     return(
@@ -422,9 +477,9 @@ permute_tau <- function(tau, labels) {
 
 # Function for spectral clustering
 # @param network a sparse adjacency matrix
-# @param n_clusters number of specified clusters
+# @param n_blocks number of specified clusters
 # @param eigenvectors_fn a function that performs eigenvector decomposition
-spec_clust_sparse <- function(network, n_clusters, eigenvectors_fn) {
+spec_clust_sparse <- function(network, n_blocks, eigenvectors_fn) {
   n <- nrow(network)
   n_vec <- ceiling(sqrt(n))
   message("Calculating eigenvectors...")
@@ -432,19 +487,19 @@ spec_clust_sparse <- function(network, n_clusters, eigenvectors_fn) {
   message("K-means clustering...")
   c <- kmeans(
     b,
-    centers = n_clusters,
+    centers = n_blocks,
     nstart = 100,
     iter.max = 20,
     algorithm = "Hartigan-Wong"
   )
   c.ind <- c$cluster
-  z_memb <- factor(c.ind, levels = 1:n_clusters)
+  z_memb <- factor(c.ind, levels = 1:n_blocks)
   z_memb
 }
 
 
 check_clusters_sparse <-
-  function(z_memb, network, n_clusters, eigenvectors_fn, min_size = 2, verbose = verbose) {
+  function(z_memb, network, n_blocks, eigenvectors_fn, min_size = 2, verbose = verbose) {
     n <- nrow(network)
     n_vec <- ceiling(sqrt(n))
     
@@ -453,7 +508,7 @@ check_clusters_sparse <-
     }
     
     b <- eigenvectors_fn(network, n_vec)
-    z_memb <- factor(z_memb, levels = 1:n_clusters)
+    z_memb <- factor(z_memb, levels = 1:n_blocks)
     
     if (verbose > 0) {
       message("Checking for bad clusters...")
@@ -463,7 +518,7 @@ check_clusters_sparse <-
     length_prev_bad_clusters <- NULL
     repeat {
       z_memb_temp <- as.vector(z_memb)
-      z_memb <- factor(z_memb_temp, levels = 1:n_clusters)
+      z_memb <- factor(z_memb_temp, levels = 1:n_blocks)
       bad_clusters <- which(table(z_memb) < min_size)
       
       if (verbose > 0) {
@@ -503,7 +558,7 @@ check_clusters_sparse <-
       for (i in 1:length(indeces)) {
         z_memb_temp[z_memb %in% indeces][temp$cluster == i] <- indeces[i]
       }
-      z_memb <- factor(z_memb_temp, levels = 1:n_clusters)
+      z_memb <- factor(z_memb_temp, levels = 1:n_blocks)
     }
     
     if (verbose > 0) {
